from __future__ import annotations

DEFAULT_REG = r"\d+$"
WORKSPACE_VID_REG = f"WSP-{DEFAULT_REG}"
PROJECT_VID_REG = f"PRJ-{DEFAULT_REG}"
PHASE_VID_REG = f"PHA-{DEFAULT_REG}"
ITERATION_VID_REG = f"ITR-{DEFAULT_REG}"
DATASET_VID_REG = f"DTS-{DEFAULT_REG}"
DATASET_VERSION_VID_REG = f"DTV-{DEFAULT_REG}"
MODEL_VID_REG = f"MDL-{DEFAULT_REG}"
MODEL_VERSION_VID_REG = f"MDV-{DEFAULT_REG}"
